// Copyright 2009,2010 Thomas Kamps anubis1@linux-ecke.de
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the7 implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

#include <QTcpSocket>
#include <QHttp>
#include <QDir>
#include <QStringList>
#include <QFile>
#include <QDomDocument>
#include <QDomElement>
#include <QXmlStreamWriter>
#include <QApplication>
#include "capiserver.h"
#include "capiserverplayer.h"

CapiServer::CapiServer(QObject* parent) : QObject(parent) {
	pID = 0;
	gID = 0;

	host    = "capid.03.testing";
	port    = "1234";
	version = "Capid 0.4";

	//Start timer to implement time dependend stuff (eg: auctions)
	timer = new QTimer(this);
	timer->start(1000);
	connect(timer, SIGNAL(timeout()), this, SLOT(tick()));

	//get all gametemplates from the games/ dir
	QDir dir("games/", "*.xml");

	QStringList files = dir.entryList();
	foreach(QString f, files) {
		QFile file(QDir::currentPath()+"/games/"+f);
		if (!file.exists()) qDebug(QString("File: "+file.fileName()+" could not be found.").toUtf8().data());
		file.open(QIODevice::ReadOnly);

		QString name = "";
		QString desc = "";
		QString type = f.replace(".xml", "");

		QString errmsg = "";
		int errline;
		int errcol;
		QDomDocument doc;
		if (doc.setContent(&file, false, &errmsg, &errline, &errcol)) {
			QDomElement cnt = doc.documentElement();
			if (cnt.nodeName() == "capigame") {
				QDomElement genData = cnt.firstChildElement("general");
				if (!genData.isNull()) {
					QDomElement elemName = genData.firstChildElement("name");
					if (!elemName.isNull()) name = elemName.text();
					QDomElement elemDesc = genData.firstChildElement("description");
					if (!elemName.isNull()) desc = elemDesc.text();
				}
			}
			CapiTemplate* tmpl = new CapiTemplate();
			tmpl->setType(type);
			tmpl->setName(name);
			tmpl->setDesc(desc);
			templates.append(tmpl);
		} else {
			qDebug(QString("Could not parse: "+f+": '"+errmsg+"' at line "+QString::number(errline)+"\n").toUtf8().data());
		}
	}

	//Start server
	serv = new QTcpServer();
	serv->listen(QHostAddress::Any, 1234);
	if (!serv->isListening()) {
		qDebug(QString("Could not start server. Probaly the port %1 is used by an other service.").arg(port).toUtf8().data());
	}
	connect(serv, SIGNAL(newConnection()), this, SLOT(manageConnection()));

	//inactive because of missing implemenation for getting config data
	//registerGator();
}

void CapiServer::tick() {
	foreach (CapiServerPlayer* p, players)
		p->tick();

	foreach (CapiServerGame* g, games)
		g->tick();
}

void CapiServer::manageConnection() {
	QTcpSocket* sock = serv->nextPendingConnection();
	CapiServerPlayer* p = new CapiServerPlayer(++pID);

	players.append(p);
	p->setSocket(sock);

	p->send("<server version=\""+version+"\" name=\"Capid\" />");
	connect(p, SIGNAL(gotMessage(int,QString)), this, SLOT(handleMessage(int,QString)));
	sock->setObjectName(QString::number(pID));
	connect(sock, SIGNAL(disconnected()), sock, SLOT(deleteLater()));
}

CapiServerPlayer* CapiServer::getPlayerById(int pid) {
	foreach(CapiServerPlayer* p, players) {
		if (p->getId() == pid) return p;
	}
	return 0;
}

CapiServerGame* CapiServer::getGameById(int gid) {
	foreach(CapiServerGame* g, games) {
		if (g->getId() == gid) return g;
	}
	return 0;
}

QString CapiServer::maskXML(QString txt) {
	QString msg = txt.replace("&", "&#38;");
	msg = msg.replace("<", "&#60;");
	msg = msg.replace(">", "&#62;");
	msg = msg.replace("=", "&#61;");
	msg = msg.replace("\"", "&#34;");
	return msg;
}

bool CapiServer::nameUsed(QString name) {
	foreach(CapiServerPlayer* p, players) {
		if (p->getName() == name) return true;
	}
	return false;
}

void CapiServer::sendToAll(QString msg, int gid) {
	foreach(CapiServerPlayer* p, players) {
		int g = -1;
		if (p->getGame() != 0) g = p->getGame()->getId();
		if ((gid == -2) || (g == gid)) p->send(msg);
	}
}

void CapiServer::deletePlayer(int pid) {
	CapiServerPlayer* p = getPlayerById(pid);
	players.removeAll(p);
	disconnect(p, SIGNAL(gotMessage(int,QString)), this, SLOT(handleMessage(int,QString)));
	delete p;
}

void CapiServer::handleMessage(int pid, QString msg) {
	CapiServerPlayer* p = getPlayerById(pid);
	if (p == 0) {
		qDebug() << QString("Server has a bug! Cannot find Player with ID: %1").arg(pid);
		return;
	}

	//command from player
	if (msg.startsWith(".")) {

		//Ablity to change name
		if (msg.startsWith(".n")) {
			QString name = msg.remove(0,2).replace("\n", "");
			//Check if name is used, and search a number to append, if needed
			if (nameUsed(name)) {
				int i = 2;
				while (nameUsed(name+QString::number(i))) {i++; }
				name = name+QString::number(i);
			}
			//inform about other players (handshake complete)
			if (p->getState() == 0) {
				QString cookie = QString::number(rand()%100000000);
				cookie += "/"+QString::number(rand()%100000000);
				p->setCookie(cookie);
				playerMap.insert(cookie, p);
				p->sendClientInfo();
				foreach (CapiServerPlayer* pl, players) {
					if (pl->getId() != pid) p->send(pl->getUpdate("name image game"));
				}
				foreach (CapiTemplate* tmpl, templates) {
					p->send(QString("<gameupdate gameid=\"-1\" gametype=\"%1\" name=\"%2\" description=\"%3\" />").arg(tmpl->getType(), tmpl->getName(), tmpl->getDesc()));
				}
				foreach (CapiServerGame* g, games) {
					p->send(g->getUpdate());
				}
			}
			p->setName(name);
			p->setState(1);

			sendToAll(p->getUpdate("name host image game"));
		} else if ((p->getState() == 1) && msg.startsWith(".pi")) {
			//Ability to chance image
			QString av = msg.remove(0,3);
			p->setAvatar(av);
			p->setState(1);
			sendToAll(p->getUpdate("image"));
		} else if ((p->getState() == 1) && msg.startsWith(".gn")) {
			//Ability to create a game
			if (p->getGame() == 0) {
				QString type = msg.remove(0,3);

				CapiServerGame* g = createGame(getTemplate(type));
				if (g != 0) {
					g->setMaster(p);
					g->addPlayer(p);
					p->setGame(g);
					sendToAll(g->getUpdate());
					sendToAll(p->getUpdate("game"));
					p->send(g->getConfigUpdate());
				} else {
					p->sendMsg("Cannot create game "+type, "error");
				}
			} else {
					p->sendMsg("You cannot create a game, when you are within a game.", "error");
			}
		} else if ((p->getState() == 1) && msg.startsWith(".gc")) {
			//Ability to change config
			bool ok;
			QStringList parts = msg.split(QString(":"));
			QString msgCid = parts.value(0);
			msgCid = msgCid.remove(0,3);
			QString msgVal = parts.value(1);
			int cid = msgCid.toInt(&ok);
			if (ok) {
				CapiServerGame* g = (CapiServerGame*) p->getGame();
				if (g != 0) {
					if (p == g->getMaster()) {
						g->setConfig(cid, msgVal);
						sendToAll(g->getConfigUpdate(cid));
					} else {
						p->sendMsg("You are not Master of the game", "error");
					}
				}
			} else {
				p->sendMsg("Invalid configid: "+msgCid, "error");
			}
		} else if ((p->getState() == 1) && msg.startsWith(".gd")) {
			//Ability to change game description
			QString desc = msg.remove(0,3);
			CapiServerGame* g = (CapiServerGame*) p->getGame();
			if (g != 0) {
				if (p->getId() == g->getMaster()->getId()) {
					g->setDesc(desc);
					sendToAll(g->getUpdate("description"));
				} else {
					p->sendMsg("You are not Master of the game", "error");
				}
			}
		} else if ((p->getState() == 1) && msg.startsWith(".gj")) {
			//Ability to join a game
			int gid = 0;
			bool ok;
			gid = msg.remove(0,3).toInt(&ok);
			if (ok) {
				CapiServerGame* g = getGameById(gid);
				if ((p->getGame() == 0) && (g != 0)) {
					if (g->getCanBeJoined()) {
						g->addPlayer(p);
						p->setGame(g);
						sendToAll(p->getUpdate("game"));
						p->send(g->getConfigUpdate());
					} else {
						p->sendMsg("You cannot join the game.", "error");
					}
				} else {
					p->sendMsg("Invalid gameid or you arer already in a game.", "error");
				}
			} else {
					p->sendMsg("Invalid gameid.", "error");
			}
		} else if ((p->getState() == 1) && msg.startsWith(".gx")) {
			//Ability to leave a game
			CapiServerGame* g = (CapiServerGame*) p->getGame();
			if (g != 0) {
				QString updateStr = g->removePlayerFromGame(p);
				p->setGame(0);
				sendToAll(p->getUpdate("game"));
				if (g->getMaster() == 0) {
					//Is the last player hs gone, delete the game
					games.removeAll(g);
					sendToAll(QString("<deletegame gameid=\"%1\" />").arg(g->getId()));
					delete g;
				} else {
					sendToAll(updateStr, g->getId());
				}
			} else {
				p->sendMsg("You are not in a game." "error");
			}
		} else if ((p->getState() == 1) && msg.startsWith(".gs")) {
			//Ablility to start a game
			CapiServerGame* g = (CapiServerGame*) p->getGame();
			if (g != 0) {
				g->startGame();
			} else {
				p->sendMsg("You are not in a game", "error");
			}
		} else if (msg.startsWith(".d")) {
			//Ability to disconnect
			//TODO: remove pid from all games
			CapiServerGame* g = (CapiServerGame*) p->getGame();
			if (g != 0) {
				g->removePlayerFromGame(p);
			}
			deletePlayer(pid);
			sendToAll(QString("<deleteplayer playerid=\"%1\" />").arg(pid));
		} else if (msg.startsWith(".R")) {
			//Ability to reconnect, if needed
			QString cookie = msg.remove(0,2);
			CapiServerPlayer* player = playerMap.value(cookie);
			if (player != 0) {
				if (player->getState() == 2) {
					//make a reconnect
					player->setState(1);
					player->setSocket(p->getSocket());
					p->setSocket(0);
					players.removeAll(p);
					playerMap.remove(p->getCookie());
					sendToAll(QString("<deleteplayer playerid=\"%1\" />").arg(p->getId()));
					delete p;
					player->sendClientInfo();
					player->send(((CapiServerGame*) player->getGame())->getFullStatus());
					player->sendMsg("Reconnected", "info");
				}
			}
		} else {
			//Commands within a game
			if (p->getGame() != 0) {
				CapiServerGame* g = (CapiServerGame*) p->getGame();
				g->command(msg, pid);
			}
		}
	} else {
		if (p->getState() != 1) return;
		//chat message
		QString name = p->getName();
		foreach(CapiServerPlayer* p, players)
			p->sendMsg(msg, "chat", pid);
	}
}

void CapiServer::registerGator() {
	QHttp* httpSock = new QHttp();

	httpSock->setHost("monopd-gator.kde.org");
	httpSock->request(QHttpRequestHeader("GET /register.php?ip="+serv->serverAddress().toString()+"&host="+host+"&port="+port+"&version="+version+"&users="+QString::number(players.count())+" HTTP/1.1\nHost: monopd-gator.kde.org\nConnection: Keep-Alive\nUser-Agent: Capid 0.3\nPragma: no-cache\nCache-control: no-cache\nAccept: text/html, text/*;q=0.9, */*;q=0.8\nAccept-Encoding: x-gzip, x-deflate, gzip, deflate\nAccept-Charset: utf-8, utf-8;q=0.5, *;q=0.5\nAccept-Language: de, en-US, en\n"));

	delete httpSock;
}

CapiTemplate* CapiServer::getTemplate(QString type) {
	foreach(CapiTemplate* tmpl, templates) {
		if (tmpl->getType() == type) return tmpl;
	}
	return 0;
}

CapiServerGame* CapiServer::createGame(CapiTemplate* tmpl) {
	if (tmpl == 0) return 0;
	QFile file(QDir::currentPath()+"/games/"+tmpl->getType()+".xml");
	file.open(QIODevice::ReadOnly);

	CapiServerGame* g = new CapiServerGame(++gID);
	g->setName(tmpl->getName());
	g->setDesc(tmpl->getDesc());
	g->setType(tmpl->getName());
	QDomDocument doc;
	doc.setContent(&file);

	g->parseGame(doc);

	games.append(g);

	return g;
}
